"""Utils."""

import contextlib
import os
from pathlib import Path
import shlex
import shutil
import subprocess
import tempfile
from typing import Any, Iterator

def pandoc() -> str:
    """Pandoc location."""
    return os.environ.get("PANDOC", "pandoc")

def grep() -> str:
    """Grep location."""
    return os.environ.get("GREP", "grep")

def check_requirements() -> bool:
    """Check if grep and pandoc are installed."""
    return bool(shutil.which(pandoc()) and shutil.which(grep()))

def sqlite_string(text: str) -> str:
    """Encode python string into sqlite string."""
    return "'{}'".format(text.replace("'", "''"))

@contextlib.contextmanager
def temporary_directory() -> Iterator[Path]:
    """Path to temporary directory."""
    with tempfile.TemporaryDirectory() as tempdir:
        yield Path(tempdir)

def run_command(cmd: str, **kwargs: Any) -> subprocess.CompletedProcess:
    """Run command with environment variables in kwargs.

    Returns stdout and stderr output.
    """
    env = os.environ.copy()
    env.update(**kwargs)
    proc = subprocess.run(shlex.split(cmd), env=env, check=False,
                          capture_output=True)
    return proc



# type: ignore
"""Functions for mocking the database."""

import sqlite3
from typing import Iterable

import pytest

from . import scan

@pytest.fixture
def mock_db() -> Iterable[sqlite3.Connection]:
    """Create an empty mock database with all the necessary tables."""
    with sqlite3.connect(":memory:") as conn:
        scan.initialize_database(conn)
        yield conn



"""Build HTML from all notes in slipbox."""

from argparse import ArgumentParser, Namespace
import os
import sqlite3
import sys

from .page import generate_complete_html
from .scan import scan, remove_outdated_files_from_database, input_files
from .scan import initialize_database
from .utils import check_requirements

def main(args: Namespace) -> None:
    """Generate HTML file from input files."""
    timestamp = 0.0
    if os.path.exists(args.database):
        timestamp = os.path.getmtime(args.database)
    with sqlite3.connect(args.database) as conn:
        initialize_database(conn)
        remove_outdated_files_from_database(conn, timestamp)
        inputs = list(input_files(conn, timestamp, args.paths, args.patterns))
        scan(conn, inputs, args.content_options, args.convert_to_data_url)
        generate_complete_html(conn, args.document_options)

if __name__ == "__main__":
    if not check_requirements():
        sys.exit("[ERROR] pandoc and grep not found.")
    parser = ArgumentParser(
        description="Generate a single-page HTML from your notes.")
    parser.add_argument("database", help="filename of sqlite3 database")
    parser.add_argument("paths", nargs='+',
                        help="list of files or directories that contain notes")
    parser.add_argument("-p", "--patterns", nargs='*', default=["*.md"],
                        help="list of glob patterns")
    parser.add_argument("-c", "--content-options", default="",
                        help="pandoc options for the content")
    parser.add_argument("-d", "--document-options", default="",
                        help="pandoc options for the final HTML output")
    parser.add_argument("--convert-to-data-url", action="store_true",
                        help="convert local image links to data URL")
    main(parser.parse_args())



"""Look for files that must be compiled."""

from itertools import chain, groupby
import fnmatch
import os
from pathlib import Path
import shlex
from sqlite3 import Connection
import sys
from typing import Iterable, List, Set, Tuple, Union

from . import utils

def initialize_database(conn: Connection) -> None:
    """Initialize database with schema.sql."""
    sql = Path(__file__).with_name("schema.sql").read_text()
    conn.executescript(sql)
    conn.commit()

def fetch_files(conn: Connection) -> Iterable[Path]:
    """Get files from the database."""
    return (Path(p) for p, in conn.execute("SELECT filename FROM Files"))

def is_recently_modified(timestamp: float, filename: Union[Path, str]) -> bool:
    """Check if file has been modified after the timestamp."""
    return os.path.exists(filename) and os.path.getmtime(filename) >= timestamp

def is_file_in_db(path: Path, conn: Connection) -> bool:
    """Check if file is recorded in the database."""
    cur = conn.cursor()
    sql = "SELECT filename FROM Files WHERE filename = ?"
    for _ in cur.execute(sql, (str(path),)):
        return True
    return False

def has_valid_pattern(filename: Path, patterns: Iterable[str]) -> bool:
    """Check if filename matches one of the patterns."""
    for pattern in patterns:
        if fnmatch.fnmatch(str(filename), pattern):
            return True
    return False

def files_in_path(path: Path) -> Iterable[Path]:
    """Recursively glob files in path."""
    if path.is_file():
        yield path
    elif path.is_dir():
        yield from filter(os.path.isfile, path.rglob('*'))

def files_in_paths(paths: Iterable[Path]) -> Set[Path]:
    """Recursively glob files in every path in paths."""
    new_files: Set[Path] = set()
    for path in paths:
        new_files = new_files.union(files_in_path(path))
    return new_files

def find_new_files(
        conn: Connection,
        paths: Iterable[Path],
        patterns: Iterable[str] = ("*.md",)
    ) -> Iterable[Path]:
    """Look for files that are not yet in the database."""
    condition = lambda x: x.is_file() and not is_file_in_db(x, conn) and \
            has_valid_pattern(x, patterns)
    return filter(condition, files_in_paths(paths))

def input_files(
        conn: Connection,
        timestamp: float,
        paths: Iterable[str],
        patterns: Iterable[str] = ("*.md",)
    ) -> Iterable[Path]:
    """Generate files that must be rescanned/processed by pandoc.

    Neighbor notes of deleted/modified files don't have to be rescanned,
    since those notes don't get affected when notes they reference to are
    removed from the database.
    """
    files = lambda: fetch_files(conn)
    modified = (p for p in files() if is_recently_modified(timestamp, p))
    new = find_new_files(conn, map(Path, paths), patterns)
    yield from filter(os.path.isfile, set(modified).union(new))

def group_by_file_extension(files: Iterable[Path]) -> Iterable[Iterable[Path]]:
    """Generate an iterator for each file extension.

    Each file with no file extension is given its own iterator.
    """
    def key(filename: Union[str, Path]) -> Tuple[str, str]:
        root, ext = os.path.splitext(filename)
        return (ext, "") if ext else ("", root)
    groups = groupby(sorted(files, key=key), key=key)
    return map(lambda g: g[1], groups)

def build_command(inputs: str, output: str, options: str = "") -> str:
    """Construct a single pandoc command to run on inputs.

    inputs is a string of filenames separated by spaces.
    Each filename must be shlex.quoted if needed.
    Return an empty string if there are no input files.
    """
    if not inputs.strip():
        return ""
    datadir = shlex.quote(str(Path(__file__).parent.resolve()))
    cmd = f"{utils.pandoc()} -Lzk.lua -Fpandoc-citeproc -Lrefs.lua --section-divs " \
            f"--data-dir {datadir} -Mlink-citations:true {options} " \
            "-o {} ".format(shlex.quote(output))
    return cmd + ' ' + inputs

def remove_outdated_files_from_database(conn: Connection, timestamp: float) -> None:
    """Remove outdated files from the database.

    This includes files that have been deleted from the file system,
    and files that have been modified recently.
    Recently modified files will be added back to the database after scanning.
    """
    files = lambda: fetch_files(conn)
    missing = (p for p in files() if not p.exists())
    modified = (p for p in files() if is_recently_modified(timestamp, p))

    cur = conn.cursor()
    cur.execute("PRAGMA foreign_keys=ON")
    for filename in chain(missing, modified):
        cur.execute("DELETE FROM Files WHERE filename = ?", (str(filename),))
    conn.commit()

def store_html_sections(conn: Connection, html: str, sources: List[Path]) -> None:
    """Insert Html and Sections entries for html and sources.

    html
    : HTML body text.
    sources
    : List of source files from which the html was generated.
    """
    if not html.strip() or not sources:
        return
    cur = conn.cursor()
    cur.execute("PRAGMA foreign_keys=ON")
    cur.execute("INSERT INTO Html(body) VALUES (?)", (html,))
    lastrowid = cur.lastrowid
    query = "SELECT DISTINCT id FROM Notes WHERE filename IN ({})".format(
        ", ".join(utils.sqlite_string(str(source)) for source in sources))
    insert = "INSERT OR IGNORE INTO Sections(note, html) VALUES (?, ?)"
    cur2 = conn.cursor()
    for row in cur.execute(query):
        cur2.execute(insert, (row[0], lastrowid))
    conn.commit()

def scan(conn: Connection, inputs: List[Path], scan_options: str, self_contained: bool) -> None:
    """Process inputs and store results in database."""
    convert_to_data_url = "1" if self_contained else ""
    for batch in group_by_file_extension(inputs):
        files = list(batch)
        scan_input_list = " ".join(shlex.quote(str(p)) for p in files)

        with utils.temporary_directory() as tempdir:
            slipbox_sql = tempdir/"slipbox.sql"
            html = tempdir/"temp.html"
            cmd = build_command(scan_input_list, str(html), scan_options)
            proc = utils.run_command(cmd, SLIPBOX_SQL=str(slipbox_sql),
                                     CONVERT_TO_DATA_URL=convert_to_data_url,
                                     GREP=utils.grep(),
                                     SCAN_INPUT_LIST=scan_input_list)
            if proc.stdout:
                print(proc.stdout.decode())
            if proc.stderr:
                print(proc.stderr.decode(), file=sys.stderr)
            if proc.returncode:
                print("Scan failed.", file=sys.stderr)
                continue
            conn.executescript(slipbox_sql.read_text())
            conn.commit()
            contents = html.read_text()
        store_html_sections(conn, contents, inputs)



# type: ignore
"""Test utils.py."""

import shlex
import shutil

import pytest

from . import utils

def test_sqlite_string():
    """Single quotes must be escaped properly."""
    assert utils.sqlite_string("''") == "''''''"
    assert utils.sqlite_string("'foo'bar'") == "'''foo''bar'''"

def test_temporary_directory():
    """temporary_directory gets deleted when the context manager exits."""
    with utils.temporary_directory() as temp:
        path = temp
        assert temp.exists()
        assert temp.is_dir()
    assert not path.exists()

@pytest.mark.skipif(not shutil.which(utils.grep()), reason="requires grep")
def test_run_command(tmp_path):
    """run_command must return stdout and stderr output."""
    first = tmp_path/"first.txt"
    second = tmp_path/"second.txt"
    first.write_text("first")
    second.write_text("second")

    proc = utils.run_command("grep first {} {}".format(
        shlex.quote(str(first)), shlex.quote(str(second))
    ))
    assert not proc.stderr
    assert not proc.returncode
    assert str(first) in proc.stdout.decode()
    assert str(second) not in proc.stdout.decode()
    proc = utils.run_command("grep second {} {}".format(
        shlex.quote(str(first)), shlex.quote(str(second))
    ))
    assert not proc.stderr
    assert str(first) not in proc.stdout.decode()
    assert str(second) in proc.stdout.decode()

@pytest.mark.skipif(not shutil.which("env"), reason="requires env")
def test_run_command_with_kwargs():
    """Keyword arguments to run_command must be used as environment variables.
    """
    proc = utils.run_command("env", ZZZZZZZZZZ="ZZZZZZZZZZ")
    assert not proc.stderr
    assert not proc.returncode
    assert "ZZZZZZZZZZ=ZZZZZZZZZZ" in proc.stdout.decode()



# type: ignore
"""Test scan.py."""

import os
import pathlib
import time

import pytest

from . import scan
from .utils import sqlite_string, check_requirements

def insert_file_script(*files: pathlib.Path):
    """Create SQL query string to insert into the Files table."""
    sql = "INSERT INTO Files (filename) VALUES ({})"
    filenames = (sqlite_string(str(p)) for p in files)
    return sql.format("), (".join(filenames))

def test_is_recently_modified(tmp_path):
    """is_recently_modified should return true iff the input file was modified
    after the specified timestamp.
    """
    before = time.time()
    path = tmp_path/"file"
    path.touch()
    os.utime(path, ns=(time.time_ns(), time.time_ns()))
    after = time.time()

    assert scan.is_recently_modified(before, path)
    assert not scan.is_recently_modified(after, path)

    os.utime(path, ns=(time.time_ns(), time.time_ns()))
    assert scan.is_recently_modified(after, path)

def test_is_file_in_db(mock_db, tmp_path):
    """Quick check for is_file_in_db."""
    conn = mock_db
    present = tmp_path/"present"
    absent = tmp_path/"absent"
    conn.executescript(insert_file_script(present.name))
    assert scan.is_file_in_db(present.name, conn)
    assert not scan.is_file_in_db(absent.name, conn)

def test_has_valid_pattern():
    """has_valid_pattern is true only for filenames that match at least one of
    the wildcard patterns."""
    patterns = ("*.md", "*.rst")
    assert scan.has_valid_pattern("a.md", patterns)
    assert scan.has_valid_pattern("a.rst", patterns)
    assert not scan.has_valid_pattern("a.tex", patterns)
    assert not scan.has_valid_pattern("a.md", (".md"))

def test_files_in_path(tmp_path):
    """files_in_path should only find files in the specified directory."""
    subdir = tmp_path/"subdir"
    inside = subdir/"inside.md"
    outside = tmp_path/"outside.md"
    nested = subdir/"nested"
    subdir.mkdir()
    inside.touch()
    outside.touch()
    nested.mkdir()

    files = [p.resolve() for p in scan.files_in_path(subdir)]
    assert inside in files
    assert outside not in files
    assert subdir not in files
    assert nested not in files

def test_files_in_path_to_file(tmp_path):
    """files_in_path should yield input if it's just a file."""
    input_file = tmp_path/"input.md"
    input_file.touch()

    files = list(scan.files_in_path(input_file))
    assert files == [input_file]

def test_find_new_files(mock_db, tmp_path):
    """find_new_files must only return existing files that aren't yet in the
    database and match the input patterns (*.md by default).
    """
    conn = mock_db
    present = tmp_path/"present.md"
    absent = tmp_path/"absent.md"
    directory = tmp_path/"directory"
    txt = tmp_path/"ignore.txt"
    present.touch()
    absent.touch()
    directory.mkdir()
    txt.touch()

    conn.executescript(insert_file_script(present))
    new_files = list(scan.find_new_files(conn, [tmp_path]))
    assert new_files == [absent]

def test_group_by_file_extension():
    """group_by_file_extension should split by file type.

    Files with no extension should be considered their own type.
    """

    files = ["a.md", "b.md", ".md", "c.tex", ".tex", ""]
    groups = list(map(list, scan.group_by_file_extension(files)))
    assert len(groups) == 5
    assert ["a.md", "b.md"] in groups
    assert [".md"] in groups
    assert ["c.tex"] in groups
    assert [".tex"] in groups
    assert [""] in groups

def test_group_by_file_extension_on_the_same_type():
    """group_by_file_extension should group files with the same type together."""
    files = [f"{c}.md" for c in "abcdefg"]
    groups = list(map(list, scan.group_by_file_extension(files)))
    assert len(groups) == 1
    assert files in groups

def test_build_command():
    """Sanity check for build_command."""
    output = "output.html"
    options = "--mathjax"
    assert not scan.build_command("  ", output, options)
    cmd = scan.build_command("cha1.md cha2.md", output, options)
    assert "cha1.md cha2.md" in cmd
    assert f"-o {output}" in cmd
    assert options in cmd

def test_remove_outdated_files_from_database(mock_db, tmp_path):
    """remove_outdated_files_from_database must remove:

    - Files that have been deleted from the file system.
    - Files that have been modified since the most recent scan.
    """
    conn = mock_db
    missing = tmp_path/"missing.md"
    modified = tmp_path/"modified.md"
    temp = tmp_path/"temp.md"
    missing.touch()
    modified.touch()
    temp.touch()

    conn.executescript(insert_file_script(missing, modified, temp))
    assert scan.is_file_in_db(missing, conn)
    assert scan.is_file_in_db(modified, conn)
    assert scan.is_file_in_db(temp, conn)

    timestamp = time.time()
    missing.unlink()
    os.utime(modified, ns=(time.time_ns(), time.time_ns()))

    scan.remove_outdated_files_from_database(conn, timestamp)

    assert not scan.is_file_in_db(missing, conn)
    assert not scan.is_file_in_db(modified, conn)
    assert scan.is_file_in_db(temp, conn)

@pytest.mark.skipif(not check_requirements(), reason="requires grep and pandoc")
def test_scan(mock_db, tmp_path):
    """Smoke test for scan."""
    input_file = tmp_path/"input.md"
    input_file.write_text("# 1 Test note\n\nHello, world!\n")
    assert not list(mock_db.execute("SELECT * FROM Html"))
    scan.scan(mock_db, [input_file], "", False)
    assert len(list(mock_db.execute("SELECT * FROM Html"))) == 1

@pytest.mark.skipif(not check_requirements(), reason="requires grep and pandoc")
def test_scan_empty_file(mock_db, tmp_path):
    """Scanned files that are empty shouldn't have entries in the database."""
    empty = tmp_path/"empty.md"
    empty.touch()
    scan.scan(mock_db, [empty], "", False)
    assert not list(mock_db.execute("SELECT * FROM Files"))
    assert not list(mock_db.execute("SELECT * FROM Notes"))
    assert not list(mock_db.execute("SELECT * FROM Tags"))
    assert not list(mock_db.execute("SELECT * FROM Links"))
    assert not list(mock_db.execute("SELECT * FROM Aliases"))
    assert not list(mock_db.execute("SELECT * FROM Sequences"))
    assert not list(mock_db.execute("SELECT * FROM Html"))
    assert not list(mock_db.execute("SELECT * FROM Sections"))
    assert not list(mock_db.execute("SELECT * FROM Bibliography"))
    assert not list(mock_db.execute("SELECT * FROM Citations"))

@pytest.mark.skipif(not check_requirements(), reason="requires grep and pandoc")
def test_scan_filenames(mock_db, tmp_path):
    """Filenames must be grepped from input files only."""
    markdown = tmp_path/"foo.md"
    skip = tmp_path/"bar.md"
    markdown.write_text("# 0 Note\n\nBody.\n")
    skip.write_text("# 0 Note\n\nBody.\n")
    scan.scan(mock_db, [markdown], "", False)

    result = list(mock_db.execute("SELECT filename FROM Notes WHERE id = 0"))
    assert len(result) == 1
    assert markdown.samefile(result[0][0])

@pytest.mark.skipif(not check_requirements(), reason="requires grep and pandoc")
def test_scan_filenames0(mock_db, tmp_path):
    """Filenames must be grepped from input files only."""
    markdown = tmp_path/"bar.md"
    skip = tmp_path/"foo.md"
    markdown.write_text("# 0 Note\n\nBody.\n")
    skip.write_text("# 0 Note\n\nBody.\n")
    scan.scan(mock_db, [markdown], "", False)

    result = list(mock_db.execute("SELECT filename FROM Notes WHERE id = 0"))
    assert len(result) == 1
    assert markdown.samefile(result[0][0])

@pytest.mark.skipif(not check_requirements(), reason="requires grep and pandoc")
def test_scan_external_sequence_links(mock_db, tmp_path):
    """Alias.owner must be equal to the root of the sequence defined by the
    alias, not to the source of the sequence link.
    """
    markdown = tmp_path/"test.md"
    markdown.write_text("""# 0 Foo

[Bar](#1 '0a')

# 1 Bar

[Bar](#1 '0b')
""")
    scan.scan(mock_db, [markdown], "", False)
    result = list(mock_db.execute("""
        SELECT id, owner, alias FROM Aliases ORDER BY alias
    """))
    assert len(result) == 3
    assert result[0] == (0, 0, '0')
    assert result[1] == (1, 0, '0a')
    assert result[2] == (1, 0, '0b')

def test_input_files_that_match_pattern(mock_db, tmp_path):
    """input_files must only return files that match the input pattern."""
    directory = tmp_path/"directory"
    markdown = tmp_path/"input.md"
    txt = tmp_path/"input.txt"
    tex = tmp_path/"ignore.tex"
    directory.mkdir()
    markdown.touch()
    txt.touch()
    tex.touch()

    conn = mock_db
    timestamp = time.time()
    patterns = ("*.md", "*.txt")
    inputs = scan.input_files(conn, timestamp, [tmp_path], patterns)
    assert sorted(inputs) == [markdown, txt]

def test_input_files_not_in_db(mock_db, tmp_path):
    """input_files must not include files already in the database."""
    new = tmp_path/"new.md"
    skip = tmp_path/"skip.md"
    new.touch()
    skip.touch()

    conn = mock_db
    conn.executescript(insert_file_script(skip))

    timestamp = time.time()
    inputs = scan.input_files(conn, timestamp, [tmp_path])
    assert list(inputs) == [new]

def test_input_files_recursive(mock_db, tmp_path):
    """input_files must find files recursively."""
    directory = tmp_path/"directory"
    new = directory/"new.md"
    directory.mkdir()
    new.touch()

    conn = mock_db
    timestamp = time.time()
    inputs = scan.input_files(conn, timestamp, [tmp_path])
    assert sorted(inputs) == [new]

def test_input_files_single_input(mock_db, tmp_path):
    """input_files must work with a single input file."""
    new = tmp_path/"new.md"
    new.touch()

    conn = mock_db
    timestamp = time.time()
    inputs = scan.input_files(conn, timestamp, [new])
    assert sorted(inputs) == [new]
